
import sys
from typing import Dict, Iterable

import requests

from icapy.config import (BASE_URL,
                          _ICA_HEADERS,
                          )
from icapy.projects import get_project_id

def get_analyses(status: str=None, max_jobs: int=5000) -> Iterable[Dict]:
    ''' find all analyses, possibly filtered by job state
    '''
    project_id = get_project_id()
    ext = f'api/projects/{project_id}/analysis:search'
    pagesize = 1000
    params = {'pageOffset': 0, 'pageSize': pagesize}
    r = requests.post(BASE_URL + ext, headers=_ICA_HEADERS, params=params)
    r.raise_for_status()
    res = r.json()
    
    # merge some statuses, since they represent stages of the same state
    states = {
        'aborted': ['aborted', 'aborting'],
        'running': ['in_progress',
                    'initializing',
                    'preparing_inputs',
                    'queued'],
        'failed': ['failed'],
        'requested': ['requested'],
        'succeeded': ['succeeded'],
    }
    
    if status is not None:
        assert status in states
    
    i = 0
    for item in res['items']:
        i += 1
        if status is not None and item['status'].lower() not in states[status]:
            continue
        yield item
        if i > max_jobs:
            break
    
    while len(res['items']) >= pagesize and i < max_jobs:
        params['pageOffset'] += pagesize
        r = requests.post(BASE_URL + ext, headers=_ICA_HEADERS, params=params)
        r.raise_for_status()
        res = r.json()
        for item in res['items']:
            i += 1
            if status is not None and item['status'].lower() not in states[status]:
                continue
            yield item

def get_analysis(analysis_id: str) -> Dict:
    ''' get details for a single analysis job
    '''
    header = {k:v for k, v in _ICA_HEADERS.items()}
    header['accept'] = 'application/vnd.illumina.v4+json'
    project_id = get_project_id()
    ext = f'api/projects/{project_id}/analyses/{analysis_id}'
    r = requests.get(BASE_URL + ext, headers=header)
    r.raise_for_status()
    return r.json()

def find_jobs(args):
    ''' command to print job info to stdout (possibly for a single status)
    '''
    statuses = ['aborted', 'running', 'failed', 'requested', 'succeeded', None]
    if args.status not in statuses:
        sys.stderr.write(f'status must be one of: {statuses}\n')
        sys.exit(1)

    if args.id is None:
        jobs = get_analyses(args.status, args.max_jobs)
    else:
        jobs = [get_analysis(args.id)]
    
    for job in jobs:
        if args.tag is not None:
            tags = set(args.tag)
            # look for any match between the supplied tags and the job tags
            if not any(bool(set(v) & tags) for v in job['tags'].values()):
                continue
            
        # jobname, job_id, time_submitted, status
        line = [job["pipeline"]["code"], 
                job["userReference"],
                job["id"],
                job['timeCreated'],
                job["status"].lower(),
                ]
        sys.stdout.write('\t'.join(line) + '\n')





from getpass import getpass
import json
import os
from pathlib import Path
import stat
from typing import Dict

import requests

BASE_URL = 'https://ica.illumina.com/ica/rest/'

def get_config_path() -> Path:
    config_folder = Path.home() / ".config" / "ica"
    config_folder.mkdir(parents=True, exist_ok=True)
    return config_folder / "config.json"

def load_config() -> Dict[str, str]:
    ''' load configuration data
    '''
    config_path = get_config_path()
    if config_path.exists():
        return json.load(open(config_path))
    else:
        return {}

def write_config(config: Dict[str, str]):
    ''' write configuration data to disk
    '''
    config_path = get_config_path()
    with open(config_path, 'wt') as handle:
        json.dump(config, handle, indent=True)
    
    # ensure user only read/write
    os.chmod(config_path, stat.S_IRUSR | stat.S_IWUSR)

def get_ica_key() -> str:
    config = load_config()
    if 'ica_api_key' not in config:
        key = getpass('provide ICA API key: ')
        # TODO: check the API key is valid before storing it
        config['ica_api_key'] = key
        write_config(config)
    
    return config['ica_api_key']

def get_tenant() -> str:
    config = load_config()
    if 'tenant' not in config:
        config['tenant'] = input('provide your ICA tenant ID: ')
        # TODO: check the tenant ID is valid
        write_config(config)
    
    return config['tenant']

def get_token() -> str:
    ''' get JWT token for authentication
    '''
    tenant = get_tenant()
    api_key = get_ica_key()
    
    ext = f'api/tokens'
    params = {'tenant': tenant}
    headers = {
        'X-API-Key': api_key,
        'accept': 'application/vnd.illumina.v3+json'
    }
    r = requests.post(BASE_URL + ext, headers=headers, params=params, stream=True)
    r.raise_for_status()
    return r.json()['token']

if "_ICA_HEADERS" not in globals():
    _ICA_HEADERS = {'x-api-key': get_ica_key(),
                'Content-Type': 'application/json'}





import io
import json
import os
from pathlib import Path
import sys
from typing import Any, Dict, Iterable, Union, Tuple

import requests

from icapy.config import (BASE_URL,
                          _ICA_HEADERS,
                          )
from icapy.projects import get_project_id

def get_data(path: Path | str) -> Iterable[Union[str, None]]:
    ''' get a data ID matching a given path in a project
    '''
    project_id = get_project_id()
    ext = f'api/projects/{project_id}/data'
    
    params = {}
    path = Path(path)
    parent = path.parent
    name = path.name
    
    params['parentFolderPath'] = str(parent)
    if not params['parentFolderPath'].endswith('/'):
        params['parentFolderPath'] += '/'
    if name != '':
        params['filename'] = name
        params['filenameMatchMode'] = 'EXACT'
    
    r = requests.get(BASE_URL + ext, params=params, headers=_ICA_HEADERS)
    r.raise_for_status()
    data = r.json()['items']
    if len(data) == 0:
        raise ValueError(f'unknown path: {path}')
    
    for item in data:
        if item['data']['details']['path'].rstrip('/') == str(path).rstrip('/'):
            yield item

def delete_file(path: str, recursive=False):
    ''' delete a file object
    '''
    if path is None:
        raise ValueError('cannot delete files without supplying a path')
    elif path == '/':
        raise ValueError('cannot delete root directory')
    
    project_id = get_project_id()
    for item in get_data(path):
        data_id = item['data']['id']
        if item['data']['details']['dataType'] == 'FOLDER' and not recursive:
            sys.stderr.write(f"cannot remove '{path}': Is a directory\n")
            sys.exit(1)
        
        ext = f'api/projects/{project_id}/data/{data_id}:delete'
        # sys.stdout.write(f'will remove {path} via {ext}')
        r = requests.post(BASE_URL + ext, headers=_ICA_HEADERS)
        r.raise_for_status()

def rm_wrapper(args):
    ''' converts CLI arguments into fucntion call for deleting file/folder
    '''
    try:
        delete_file(args.PATH, args.recursive)
    except ValueError as err:
        sys.stderr.write(err.args[0] + '\n')
        sys.exit(1)

def mv(old_path: str, new_path: str):
    raise NotImplementedError

def get_object_details(item: Dict[str, Any]):
    ''' get details of an object
    '''
    return {
        'name': item['data']['details']['name'],
        'path': item['data']['details']['path'],
        'created_date': item['data']['details']['timeCreated'],
        'size': item['data']['details']['fileSizeInBytes'],
        'id': item['data']['id'],
    }

def list_files(path: str, pattern: str=None) -> Iterable[Dict[str, Any]]:
    ''' list details for file or folder contents
    '''
    
    project_id = get_project_id()
    pagesize = 1000
    header = {k: v for k, v in _ICA_HEADERS.items()}
    
    data = [{'data': {'id': None, 'details': {'dataType': None}}}]
    if str(path) != '/' and path is not None:
        data = get_data(path)
    
    for item in data:
        if item['data']['details']['dataType'] == 'FILE':
            # if the path corresponds to a single file, yield that without
            # making extra checks
            yield get_object_details(item)
            continue
        
        data_id = item['data']['id']
        params = {'pageOffset': 0, 'pageSize': pagesize}
        if data_id is None:
            params['parentFolderPath'] = '/'
        else:
            params['parentFolderId'] = data_id
        
        if pattern is not None:
            params['filename'] = pattern
            params['filenameMatchMode'] = 'FUZZY'
        
        ext = f'api/projects/{project_id}/data'
        r = requests.get(BASE_URL + ext, params=params, headers=header)
        r.raise_for_status()
        res = r.json()
        
        for item in res['items']:
            yield get_object_details(item)
        
        while len(res['items']) >= pagesize:
            params['pageOffset'] += pagesize
            r = requests.get(BASE_URL + ext, headers=header, params=params)
            r.raise_for_status()
            res = r.json()
            for item in res['items']:
                yield get_object_details(item)

def format_size(size):
    ''' convert filesize in bytes to human-readable form (e.g. 3.5G)
    '''
    codes = ' KMGTPE'
    factor = 0
    while factor <= len(codes):
        scaled = size / (1024 ** factor)
        if scaled < 1024:
            scaled = f'{scaled:.1f}' if 0 < scaled < 10 else f'{int(round(scaled))}'
            return f'{scaled}{codes[factor].strip()}'
        factor += 1
    
    scaled = f'{scaled:.1f}' if 0 < scaled < 10 else f'{int(round(scaled))}'
    return f'{scaled}{codes[factor].strip()}'

def ls_wrapper(args):
    ''' converts the CLI args to a function call for listiong file/folder contents
    '''
    if args.FILE is not None and not str(args.FILE).startswith('/'):
        sys.stderr.write(f'filepath must begin with "/": {args.FILE}')
        sys.exit(1)
    
    try:
        for item in list_files(args.FILE, args.PATTERN):
            if not args.all and item["name"].startswith('.'):
                continue
            line = [item["path"]]
            if args.l:
                line += [item["id"].lower(), format_size(item["size"]), item['created_date']]
            sys.stdout.write('\t'.join(line) + '\n')
    except ValueError:
        # only raises ValueError if the file/folder does not exist. If you look
        # for an empty folder, this will not be used.
        sys.stderr.write(f'cannot access {args.FILE}: No such file or directory')
        sys.exit(1)

def get_file(path: str):
    ''' get a file contents (streamed in chunks of 1600 bytes)
    '''
    project_id = get_project_id()
    data = get_data(path)
    if data is None:
        raise ValueError(f'cannot access data at {path}')
    data = list(data)
    if len(data) > 1:
        raise ValueError(f'too many matches at {path}')
    
    data_id = data[0]['data']['id']
    
    ext = f'api/projects/{project_id}/data/{data_id}:createDownloadUrl'
    r = requests.post(BASE_URL + ext, headers=_ICA_HEADERS)
    r.raise_for_status()
    
    url = r.json()['url']
    r = requests.get(url, stream=True)
    r.raise_for_status()
    return r.iter_content(1600)

def download_file(args):
    ''' download a file from ICA storage
    '''
    try:
        # get_file returns a stream of bytes, which we simply write to stdout
        for x in get_file(args.PATH):
            sys.stdout.buffer.write(x)
    except ValueError as err:
        sys.stderr.write(err.args[0] + '\n')
        sys.exit(1)
    except BrokenPipeError:
        devnull = os.open(os.devnull, os.O_WRONLY)
        os.dup2(devnull, sys.stdout.fileno())
        sys.exit(0)

def get_upload_folder(path: Path | str | None, overwrite=False) -> Tuple[str, str]:
    ''' find the details of a folder to upload data to
    
    Args:
        path: path to upload data to. This could be either full path with extension,
            a path to a folder, or 
            None                           - upload to root directory
            /FOLDER/SUBFOLDER              - need to find details of subfolder
            /FOLDER/SUBFOLDER/BASENAME.EXT - need to find details of subfolder
            
            These folders/paths may or may not exist already.
    
    Returns:
        a tuple of (folder_id, folder_path)
    '''
    if path is None:
        sys.stderr.write('uploading to root folder')
        data = next(get_data('/'))
        return data['data']['id'], data['data']['details']['path']
    
    try:
        data = next(get_data(path))
        datatype = data['data']['details']['dataType']
        if datatype == 'FOLDER':
            return data['data']['id'], data['data']['details']['path']
        elif datatype == 'FILE':
            if overwrite:
                delete_file(data['data']['details']['path'])
                raise ValueError
                # raise NotImplementedError('add code for overwriting')
            else:
                # can't raise ValueError here, otherwise the catch handler would
                # try the parent folder
                raise RuntimeError(f'would overwrite existing file: {path}, use -f/--force to overwrite')
        else:
            raise TypeError(f'unknown datatype: {datatype}')
    except ValueError:
        # if we've tried a full path that does not exist, get the parent folder
        data = next(get_data(Path(path).parent))
        datatype = data['data']['details']['dataType']
        if datatype == 'FOLDER':
            return data['data']['id'], data['data']['details']['path']
        else:
            raise ValueError(f'unknown folder: {path}')
    except RuntimeError as e:
        raise ValueError(*e.args)

def get_upload_name(infile, destination: Path | str, folder: Path | str):
    ''' determine what the name of the file will be on ICA storage
    '''
    name = Path(destination).relative_to(Path(folder))
    if str(name) != '.':
        return str(name)
    elif type(infile) == str:
        return Path(infile).name
    else:
        raise ValueError(f'cannot determine filename to save as from --path ({destination})')

def upload_file(infile: str | io.BufferedReader | bytes, destination: Path, overwrite=False):
    ''' upload a file to ICA
    
    Args:
        infile: path to a file to upload, or a file handle for reading (e.g. open 
            file or sys.stdin), or byte sequence to upload
        destination: path to save data to on ICA. Can either be a folder (in
            which case the written file uses have the infile name), or a 
            complete file path.
        overwrite: whether to overwrite if file exists already
    '''
    project_id = get_project_id()
    folder_id, folder_path = get_upload_folder(destination, overwrite)
    upload_name = get_upload_name(infile, destination, folder_path)
    
    body = {'name': upload_name,
            'folderId': folder_id
            }
    body = json.dumps(body)
    
    ext = f'api/projects/{project_id}/data:createFileWithUploadUrl'
    try:
        r = requests.post(BASE_URL + ext, data=body, headers=_ICA_HEADERS)
        r.raise_for_status()
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 409:
            if overwrite:
                delete_file(upload_name)
                r = requests.post(BASE_URL + ext, data=body, headers=_ICA_HEADERS)
                r.raise_for_status()
            else:
                raise ValueError(f'error: file already exists at {folder_path}{upload_name}')
        raise e
    
    if type(infile) == str:
        infile = open(infile, 'rb')

    url = r.json()['uploadUrl']
    try:
        r = requests.put(url, data=infile, stream=True, headers=_ICA_HEADERS)
        r.raise_for_status()
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 409:
            raise ValueError
        raise e

def upload_wrapper(args):
    ''' upload a file to ICA storage
    '''
    infile = args.INFILE
    if infile is None:
        # read stdin into memory. This might use too much memory with large inputs.
        # TODO: figure out how to use sys.stdin directly, upload stdin directly
        # TODO: via stream. Currently using sys.stdin raises an error. Fixing this
        # TODO: would permit simple moving/coping of files by downloading and 
        # TODO: re-uploading.
        infile = sys.stdin.buffer.read()
    
    try:
        # get_file returns a stream of bytes, which we simply write to stdout
        upload_file(infile, args.path, args.force)
    except ValueError as err:
        sys.stderr.write(err.args[0] + '\n')
        sys.exit(1)





import sys
from typing import Dict, Iterable

import requests

from icapy.config import (BASE_URL,
                          load_config,
                          write_config,
                          get_ica_key,
                          get_tenant,
                          _ICA_HEADERS,
                          )

def get_projects() -> Iterable[Dict]:
    ''' find all projects
    '''
    ext = 'api/projects'
    pagesize = 10
    params = {'pageOffset': 0, 'pageSize': pagesize}
    r = requests.get(BASE_URL + ext, headers=_ICA_HEADERS, params=params)
    r.raise_for_status()
    res = r.json()
    
    for item in res['items']:
        yield item

    while len(res['items']) >= pagesize:
        params['pageOffset'] += pagesize
        r = requests.get(BASE_URL + ext, headers=_ICA_HEADERS, params=params)
        r.raise_for_status()
        res = r.json()
        for item in res['items']:
            yield item

def list_projects(*args):
    ''' command to list projects to stdout
    '''
    for project in get_projects():
        # project name and ID
        line = [project["name"],
                project["id"],
                ]
        sys.stdout.write('\t'.join(line) + '\n')

def set_default_project(*args):
    ''' set the default project
    '''
    projects = [(x['name'], x['id']) for x in get_projects()]
    names, ids = zip(*projects)
    if len(names) == 0:
        sys.stderr.write('no projects available to select from\n')
        sys.exit(1)
        
    if len(names) == 1:
        sys.stderr.write(f'one project available, setting as default ({names[0]})')
        selection = 0
    else:
        sys.stderr.write("set default project:\n")
        for i, name in enumerate(names):
            sys.stderr.write(f' {i + 1}: {name}\n')
        while True:
            selection = input("provide project number to set as default: ")
            if not str.isnumeric(selection):
                sys.stderr.write('must provide integer\n')
                continue
            selection = int(selection) - 1
            if selection < 0 or selection > len(ids) - 1:
                sys.stderr.write('choice must be one of the displayed options\n')
                for i, name in enumerate(names):
                    sys.stderr.write(f' {i + 1}: {name}\n')
                continue
            
            break
    
    config = load_config()
    if 'tenant' not in config:
        config['tenant'] = get_tenant()
    
    if 'ica_api_key' not in config:
        config['ica_api_key'] = get_ica_key()
    
    config['ica_project_id'] = ids[selection]
    config['ica_project_name'] = names[selection]
    
    write_config(config)

def set_default_project(*args):
    ''' set the default project
    '''
    projects = [(x['name'], x['id']) for x in get_projects()]
    names, ids = zip(*projects)
    if len(names) == 0:
        sys.stderr.write('no projects available to select from\n')
        sys.exit(1)
        
    if len(names) == 1:
        sys.stderr.write(f'one project available, setting as default ({names[0]})')
        selection = 0
    else:
        sys.stderr.write("set default project:\n")
        for i, name in enumerate(names):
            sys.stderr.write(f' {i + 1}: {name}\n')
        while True:
            selection = input("provide project number to set as default: ")
            if not str.isnumeric(selection):
                sys.stderr.write('must provide integer\n')
                continue
            selection = int(selection) - 1
            if selection < 0 or selection > len(ids) - 1:
                sys.stderr.write('choice must be one of the displayed options\n')
                for i, name in enumerate(names):
                    sys.stderr.write(f' {i + 1}: {name}\n')
                continue
            
            break
    
    config = load_config()
    if 'tenant' not in config:
        config['tenant'] = get_tenant()
    
    if 'ica_api_key' not in config:
        config['ica_api_key'] = get_ica_key()
    
    config['ica_project_id'] = ids[selection]
    config['ica_project_name'] = names[selection]
    
    write_config(config)

def get_project_id() -> str:
    config = load_config()
    if 'ica_project_id' not in config:
        set_default_project()
        config = load_config()
    
    return config['ica_project_id']





import argparse
import os
from pathlib import Path
import sys

from icapy.data import (ls_wrapper,
                        download_file,
                        upload_wrapper,
                        rm_wrapper,
                        )
from icapy.jobs import find_jobs
from icapy.projects import set_default_project

def CLI():
    ''' small CLI application to run ICA commands
    '''
    parser = argparse.ArgumentParser(description="small CLI application to run ICA commands")
    subparsers = parser.add_subparsers()
    
    login = subparsers.add_parser('select', help="set default project")
    login.set_defaults(func=set_default_project)
    
    ls = subparsers.add_parser('ls', help="list file/folder details")
    ls.add_argument('FILE', nargs='?', type=Path, help='path to file/folder')
    ls.add_argument('--pattern', nargs='?', help='search pattern (if ls-ing a folder)')
    ls.add_argument('-l', default=False, 
                    action='store_true', help='use a long listing format')
    ls.add_argument('-a', '--all', default=False, 
                    action='store_true', help='show hidden files')
    ls.set_defaults(func=ls_wrapper)
    
    download = subparsers.add_parser('download', help="download file")
    download.add_argument('PATH', type=Path, help='path to file')
    download.set_defaults(func=download_file)
    
    upload = subparsers.add_parser('upload', help="upload file")
    upload.add_argument('INFILE', nargs='?', help='path to local file, tries stdin if not used')
    upload.add_argument('--path', type=Path, help='path to destination file (full path or folder)')
    upload.add_argument('-f', '--force', default=False, action='store_true',
                        help='whether to overwrite if detination file exists')
    upload.set_defaults(func=upload_wrapper)
    
    rm = subparsers.add_parser('rm', help="remove (unlink) the FILE(s)")
    rm.add_argument('PATH', type=Path, help='path to file/folder')
    rm.add_argument('-r','--recursive', default=False, action='store_true',
                    help='remove directories and their contents')
    rm.set_defaults(func=rm_wrapper)
    
    jobs = subparsers.add_parser('jobs', 
                                 help="find analyses/jobs",
                                 description="Jobs are sorted by most " \
                                              "recently submitted first")
    jobs.add_argument('status', nargs='?',
                      help='optionally filter by job status. If used, must be ' \
                           'one of aborted, running, failed, requested, succeeded.')
    jobs.add_argument('--id',
                      help='To get details of a specific job. This overrides ' \
                           'any status requirement.')
    jobs.add_argument('--tag', nargs='*', help='tag to filter on.')
    jobs.add_argument('--max-jobs', type=int, default=5000,
                      help='Number of jobs to get (default=5000).')
    jobs.set_defaults(func=find_jobs)
    
    args = parser.parse_args()
    if not hasattr(args, 'func'):
        parser.print_help()
    else:
        try:
            args.func(args)
        except (KeyboardInterrupt, BrokenPipeError):
            devnull = os.open(os.devnull, os.O_WRONLY)
            os.dup2(devnull, sys.stdout.fileno())
            sys.exit(0)
